import chemexp
from chemprop.utils import load_checkpoint

explainer = chemexp.PathExplainer(load_checkpoint("fusion_db_1.pt"))
exp = explainer.explain_molecule("COC(=N)Cc1ccccc1", n=0)
cpwhy.exp_to_png(exp, "/tmp/mol.png")



from chemprop.features.featurization import MolGraph, BatchMolGraph
from chemprop.models import MoleculeModel
from .explained_mol_graph import ExplainedMolGraph
from .path_explainer_torch import PathExplainerTorch
from copy import deepcopy
import torch
import numpy as np


class PathExplainer:
    # +1 is to include room for common values, as mentioned in chemprop.features.featurization, line 31
    feature_sizes = [
        100 + 1, # atom_type
        6 + 1,   # #bonds
        5 + 1,   # formal_charge
        4 + 1,   # charilaty
        5 + 1,   # #Hs
        5 + 1,   # hybridization
        1,       # aromaticity
        1        # atomic_mass
    ]

    possible_atoms = [[i%101, i%7, i%6, i%5, i%6, i%6, i%2, i/100] for i in range(101)]
    possible_bonds = [[i%5, i%2, i%2, i%7] for i in range(7)]

    def __init__(self, model: MoleculeModel, mol: str=None):
        self.model = model
        self.model.eval() # Trigger evaluation mode
        if mol is not None:
            self._set_molecule(mol)
        else:
            self.mol = None

    def _set_molecule(self, mol: str):
        mol = MolGraph(mol)
        self.mol = mol
        self.batch_mol = BatchMolGraph([mol])
        atom_size = len(mol.f_atoms[0])
        atoms = np.concatenate(mol.f_atoms)
        bonds = np.concatenate(list(map(lambda x: x[atom_size:], mol.f_bonds)))
        atom_size = len(mol.f_atoms[0])
        bond_size = len(mol.f_bonds[0]) - atom_size
        self.input = torch.Tensor(np.concatenate([np.zeros(atom_size), atoms, np.zeros(bond_size), bonds]))

    def _molecule_to_compact(self, mol: str):
        mol = MolGraph(mol)
        input_features = []
        self.categorical = []
        self.atom_slices = [0]
        values = []
        i = 0
        # atoms to array
        for f_atom in mol.f_atoms:
            values = []
            j = 0
            for size in self.feature_sizes:
                values.append(f_atom[j:j+size].index(1))
                j += size
            values.extend(f_atom[j:])
            input_features += values
            self.categorical += list(range(i, i+len(values)-1))
            i += len(values)
            self.atom_slices.append(i)
        self.atom_size = len(values)
        # bonds to array
        self.bond_slices = [i]
        for f_bond in mol.f_bonds:
            f_bond = f_bond[-14:]
            values = []
            values.append(f_bond[:5].index(1))  # bond type
            values.extend(f_bond[5:7])          # conjugated / in ring
            values.append(f_bond[7:].index(1))  # stereo (none, any, E/Z, or cis/trans)
            input_features += values
            self.categorical += list(range(i, i+4))
            i += 4
            self.bond_slices.append(i)
        self.bond_size = len(values)
        return np.array(input_features, dtype="float32")

    def _tensor_to_batch_mol(self, array) -> BatchMolGraph:
        # parsing atoms features
        atom_size = len(self.mol.f_atoms[0])
        bond_size = len(self.mol.f_bonds[0]) - atom_size
        atoms_length = (self.mol.n_atoms + 1) * atom_size
        f_atoms = array[:atoms_length].reshape((-1, atom_size))
        f_bonds = array[atoms_length:].reshape((-1, bond_size))
        # creating a copy of the MolGraph and change its features
        batch_mol = deepcopy(self.batch_mol)
        f_bonds = torch.cat([f_atoms[batch_mol.b2a], f_bonds], dim=1)
        batch_mol.f_atoms = f_atoms
        batch_mol.f_bonds = f_bonds
        return batch_mol

    def _feature_to_indexes(self, feature: int):
        atom_size = self.batch_mol.f_atoms.shape[1]
        bond_size = self.batch_mol.f_bonds.shape[1]
        if feature < len(self.feature_sizes):
            i = sum(self.feature_sizes[:feature])
            j = i + self.feature_sizes[feature]
            return [atom_size*k + i + self.batch_mol.f_atoms[k,i:j].argmax() for k in range(1, self.mol.n_atoms+1)]
        else:
            feature -= len(self.feature_sizes)
            feature_sizes = [5, 1, 1, 7]
            i = atom_size + sum(feature_sizes[:feature])
            j = i + feature_sizes[feature]
            return [i + bond[i:j].argmax() for bond in self.batch_mol.f_bonds[1:]]

    def predict_proba(self, inputs, softmax_output=False):
        """Compute the model's prediction for arrays inputs

        inputs: list of arrays, containing features for the current molecule
        softmax_output: flag to apply a softmax. If False the original linear output is returned

        Returns a numpy array of predictions, of shape (len(inputs), nb_classes)
        """
        if self.mol is None:
            raise AttributeError("You must set a molecule first (set_molecule method).")
        # converting arrays to MolGraphs
        mols = list(map(self._tensor_to_batch_mol, inputs))
        # getting the model predictions for the molecules
        output = self.model(mols)
        output = torch.Tensor.cpu(output)
        #if softmax_output:
        #    output = softmax(output, dim=1)
        #else:
        #    output=functional.normalize(output, p=1, dim=1)
        return output

    def explain_molecule(self, mol: str, n=1) -> ExplainedMolGraph:
        self._set_molecule(mol)
        explainer = PathExplainerTorch(lambda batch: self.predict_proba(batch)[:,n:n+1])
        input_tensor = self.input.reshape((1, -1))
        input_tensor.requires_grad_(True)
        baseline = torch.zeros_like(input_tensor)
        attributions = explainer.attributions(
            input_tensor=input_tensor,
            baseline=baseline,
            num_samples=256,
            use_expectation=True
        )
        atom_size = len(self.mol.f_atoms[0])
        bond_size = len(self.mol.f_bonds[0]) - atom_size
        atoms_length = (self.mol.n_atoms + 1) * atom_size
        atoms_att = attributions[0][:atoms_length].detach().split(atom_size)
        bonds_att = attributions[0][atoms_length:].detach().split(bond_size)
        atoms_att = atoms_att[1:]
        bonds_att = bonds_att[1:]
        f_atoms = []
        indexes = [0] + np.cumsum(self.feature_sizes).tolist()
        for atom in atoms_att:
            exp = np.zeros(len(self.feature_sizes))
            for i, (j, k) in enumerate(zip(indexes[:-1], indexes[1:])):
                exp[i] = atom[j:k].sum()
            f_atoms.append(exp)
        f_bonds = []
        for bond in bonds_att:
            bond = bond[-14:]
            exp = np.zeros(4)
            exp[0] = bond[:5].sum()  # bond type
            exp[1:3] = bond[5:7]     # conjugated / in ring
            exp[3] = bond[7:].sum()  # stereo (none, any, E/Z, or cis/trans)
            f_bonds.append(exp)
        exp = ExplainedMolGraph(mol, f_atoms, f_bonds)
        return exp

    def explain_interactions(self, mol: str, feature1: int, feature2: int, n=1) -> ExplainedMolGraph:
        self._set_molecule(mol)
        explainer = PathExplainerTorch(lambda batch: self.predict_proba(batch)[:,n:n+1])
        input_tensor = self.input.reshape((1, -1))
        input_tensor.requires_grad_(True)
        baseline = torch.zeros_like(input_tensor)
        featuresA = self._feature_to_indexes(feature1)
        featuresB = self._feature_to_indexes(feature2)
        interactions = explainer.single_interactions(
            input_tensor=input_tensor,
            baseline=baseline,
            featuresA=featuresA,
            featuresB=featuresB,
            num_samples=256,
            use_expectation=True,
            verbose=True
        )[0].detach().unsqueeze(-1).numpy()
        if feature1 < len(self.feature_sizes):
            f_atoms = [interactions[i,i] for i in range(len(featuresA))]
            f_bonds = []
        else:
            f_atoms = []
            f_bonds = [interactions[i,i] for i in range(len(featuresA))]
        return ExplainedMolGraph(mol, f_atoms, f_bonds)

    def __call__(self, *args, **kwargs):
        return self.predict_proba(*args, **kwargs)



import sys
from .server import server


if __name__ == "__main__":
    print(sys.argv)
    if len(sys.argv) != 2:
        print("Usage: chemexp <path>\n",
              "<path> must contain chemprop models/checkpoints (.pt files)",
              file=sys.stderr)
        exit(1)
    else:
        server.set_models_dir(sys.argv[1])
        server.run(debug=True)



#!/usr/bin/env python
import functools
import operator
import torch
from torch.autograd import grad
import numpy as np
from tqdm import *

def gather_nd(params, indices):
    """
    Args:
        params: Tensor to index
        indices: k-dimension tensor of integers. 
    Returns:
        output: 1-dimensional tensor of elements of ``params``, where
            output[i] = params[i][indices[i]]
            
            params   indices   output

            1 2       1 1       4
            3 4       2 0 ----> 5
            5 6       0 0       1
    """
    max_value = functools.reduce(operator.mul, list(params.size())) - 1
    indices = indices.t().long()
    ndim = indices.size(0)
    idx = torch.zeros_like(indices[0]).long()
    m = 1

    for i in range(ndim)[::-1]:
        idx += indices[i]*m
        m *= params.size(i)

    idx[idx < 0] = 0
    idx[idx > max_value] = 0
    return torch.take(params, idx)

class PathExplainerTorch(object):
    def __init__(self, model):
        self.model = model
        return
    
    def _get_ref_tensor(self,baseline,batch_size,num_samples):
        number_to_draw = num_samples * batch_size
        replace = baseline.shape[0] < number_to_draw
        sample_indices = np.random.choice(baseline.shape[0],
                                          size=number_to_draw,
                                          replace=replace)
        ref_tensor = baseline[sample_indices,:]
        
        return ref_tensor

    def _get_samples_input(self, input_tensor, baseline, 
                           num_samples, use_expectation):
        '''
        calculate interpolation points
        Args:
            input_tensor: Tensor of shape (batch, ...), where ... indicates
                          the input dimensions. 
            reference_tensor: A tensor of shape (batch, k, ...) where ... 
                indicates dimensions, and k represents the number of background 
                reference samples to draw per input in the batch.
        Returns: 
            samples_input: A tensor of shape (batch, k, ...) with the 
                interpolated points between input and ref.
            samples_delta: A tensor of shape (batch, 1, ...) with the
                difference between input and reference for each sample
        '''
        input_dims = list(input_tensor.size())[1:]
        num_input_dims = len(input_dims)
        batch_size = input_tensor.size()[0]
        
        if use_expectation:
            reference_tensor = self._get_ref_tensor(baseline,batch_size,num_samples)
            shape = reference_tensor.shape
            reference_tensor = reference_tensor.view(
                    batch_size, 
                    num_samples, 
                    *(shape[1:]))
            
            # Grab a [batch_size, k]-sized interpolation sample
            t_tensor = torch.FloatTensor(batch_size, num_samples).uniform_(0,1).to(reference_tensor.device)
            shape = [batch_size, num_samples] + [1] * num_input_dims
            interp_coef = t_tensor.view(*shape)

            # Evaluate the end points
            end_point_ref = (1.0 - interp_coef) * reference_tensor

            input_expand_mult = input_tensor.unsqueeze(1)
            end_point_input = interp_coef * input_expand_mult

            # Affine Combine
            samples_input = end_point_input + end_point_ref
            
        else:
            batch_size = input_tensor.size()[0]
            input_expand = input_tensor.unsqueeze(1)
            reps = np.ones(len(baseline.shape)).astype(int)
            reps[0] = batch_size
            reference_tensor = baseline.repeat(list(reps)).unsqueeze(1)
#             reference_tensor = torch.as_tensor(sampled_baseline).unsqueeze(1).to(baseline.device)
            scaled_inputs = [reference_tensor + (float(i)/num_samples)*(input_expand - reference_tensor) \
                             for i in range(0,num_samples+1)]
            samples_input = torch.cat(scaled_inputs,dim=1)
        
        samples_delta = self._get_samples_delta(input_tensor, reference_tensor)
        samples_delta = samples_delta.to(samples_input.device)
        
        return samples_input, samples_delta
    
    def _get_samples_delta(self, input_tensor, reference_tensor):
        input_expand_mult = input_tensor.unsqueeze(1)
        sd = input_expand_mult - reference_tensor
        return sd
    
    def _get_grads(self, samples_input, output_indices=None):

        grad_tensor = torch.zeros(samples_input.shape).float().to(samples_input.device)
        
        k_ = samples_input.shape[1]

        for i in range(k_):
            particular_slice = samples_input[:,i]
            batch_output = self.model(particular_slice)
            # should check that users pass in sparse labels
            # Only look at the user-specified label
            if batch_output.size(1) > 1:
                sample_indices = torch.arange(0,batch_output.size(0)).to(samples_input.device)
                indices_tensor = torch.cat([
                        sample_indices.unsqueeze(1), 
                        output_indices.unsqueeze(1)], dim=1)
                batch_output = gather_nd(batch_output, indices_tensor)

            model_grads = grad(
                    outputs=batch_output,
                    inputs=particular_slice,
                    grad_outputs=torch.ones_like(batch_output).to(samples_input.device),
                    create_graph=True)
            grad_tensor[:,i,:] = model_grads[0]
        return grad_tensor
           
    def attributions(self, input_tensor, baseline,
                     num_samples = 50, use_expectation=True, 
                     output_indices=None):
        """
        Calculate either Expected or Integrated Gradients approximation of 
        Aumann-Shapley values for the sample ``input_tensor``.

        Args:
            model (torch.nn.Module): Pytorch neural network model for which the
                output should be explained.
            input_tensor (torch.Tensor): Pytorch tensor representing the input
                to be explained.
            baseline (torch.Tensor): Pytorch tensor representing the baseline.
                If use_expectation is true, then baseline should be shape
                (num_refs, ...) where ... indicates the dimensionality
                of the input. Otherwise, baseline should be shape (1, ...).
            output_indices (optional, default=None): For multi-class prediction
        """
        equal_dims = baseline.shape[1:] == input_tensor.shape[1:]
        almost_equal_dims = baseline.shape == input_tensor.shape[1:]
        
        dev = input_tensor.device
        baseline = baseline.to(dev)
        
        input_tensor.requires_grad_ = True
        
        if use_expectation and not equal_dims:
            raise ValueError('baseline should be shape (num_refs, ...) \
                              where ... indicates the dimensionality   \
                              of the input')
            
        if not use_expectation and baseline.shape[0] != 1:
            if almost_equal_dims:
                baseline = baseline.unsqueeze(0)
            else:
                raise ValueError('baseline should be shape (...)           \
                                  where ... indicates the dimensionality   \
                                  of the input')
        
        samples_input, samples_delta = self._get_samples_input(input_tensor, baseline,
                                                               num_samples, use_expectation)
        grad_tensor = self._get_grads(samples_input, output_indices)
        mult_grads = samples_delta * grad_tensor
        attributions = mult_grads.mean(1)
        
        return attributions
    
    def interactions(self, input_tensor, baseline,
                     num_samples=50, use_expectation=True,
                     output_indices=None, interaction_index=None,
                     verbose=True):
        """
        samples_input: A tensor of shape (batch, k, features)
        ig_tensor: also size (batch, k, features), but contains IG values
        
        """
        
        if len(input_tensor.shape) != 2:
            raise ValueError('PyTorch Explainer only supports ' + \
                             'interaction for 2D input tensors!')
        
        equal_dims = baseline.shape[1:] == input_tensor.shape[1:]
        almost_equal_dims = baseline.shape == input_tensor.shape[1:]
        
        if use_expectation and not equal_dims:
            raise ValueError('baseline should be shape (num_refs, ...) \
                              where ... indicates the dimensionality   \
                              of the input')
            
        if not use_expectation and baseline.shape[0] != 1:
            if almost_equal_dims:
                baseline = baseline.unsqueeze(0)
            else:
                raise ValueError('baseline should be shape (...)           \
                                  where ... indicates the dimensionality   \
                                  of the input')
        
        inner_loop_nsamples = int(round(np.sqrt(num_samples)))
        
        samples_input, samples_delta = self._get_samples_input(input_tensor, baseline,
                                                               inner_loop_nsamples, use_expectation)
        
        if interaction_index is not None:
            interaction_mult_tensor = torch.zeros([input_tensor.size(0), samples_input.size(1), input_tensor.size(1)])
        else:
            interaction_mult_tensor = torch.zeros([input_tensor.size(0), samples_input.size(1), 
                                                   input_tensor.size(1), input_tensor.size(1)])
            
        ig_tensor = torch.zeros(samples_input.shape).float()
        
        if use_expectation:
            loop_num = inner_loop_nsamples
        else:
            loop_num = inner_loop_nsamples + 1
        
        if verbose:
            iterable = tqdm(range(loop_num))
        else:
            iterable = range(loop_num)
        
        for i in iterable:
            
            particular_slice = samples_input[:,i]
            ig_tensor[:,i,:] = self.attributions(particular_slice, baseline,
                                                 num_samples=inner_loop_nsamples, use_expectation=use_expectation,
                                                 output_indices=output_indices)
            
            if interaction_index is not None:
                second_grads = grad(
                        outputs=ig_tensor[:,i,interaction_index],
                        inputs=particular_slice,
                        grad_outputs=torch.ones_like(ig_tensor[:,i,interaction_index]),
                        create_graph=True)[0]
                interaction_mult_tensor[:,i,:] = second_grads

            else:
                for feature in range(input_tensor.size(1)):
                    second_grads = grad(
                        outputs=ig_tensor[:,i,feature],
                        inputs=particular_slice,
                        grad_outputs=torch.ones_like(ig_tensor[:,i,feature]),
                        create_graph=True)[0]
                    interaction_mult_tensor[:,i,feature,:] = second_grads

        interaction_mult_tensor = interaction_mult_tensor.to(samples_delta.device)
        if interaction_index is not None:
            interaction_tensor = interaction_mult_tensor * samples_delta
        else:
            interaction_tensor = interaction_mult_tensor * samples_delta.unsqueeze(2)
        interactions = interaction_tensor.mean(1)
        
        return interactions

    def single_interactions(self, input_tensor, baseline, featuresA, featuresB,
                            num_samples=50, use_expectation=True,
                            output_indices=None, verbose=True):

        if len(input_tensor.shape) != 2:
            raise ValueError('PyTorch Explainer only supports ' + \
                             'interaction for 2D input tensors!')

        equal_dims = baseline.shape[1:] == input_tensor.shape[1:]
        almost_equal_dims = baseline.shape == input_tensor.shape[1:]

        if use_expectation and not equal_dims:
            raise ValueError('baseline should be shape (num_refs, ...) \
                              where ... indicates the dimensionality   \
                              of the input')

        if not use_expectation and baseline.shape[0] != 1:
            if almost_equal_dims:
                baseline = baseline.unsqueeze(0)
            else:
                raise ValueError('baseline should be shape (...)           \
                                  where ... indicates the dimensionality   \
                                  of the input')

        features = list(featuresA) + list(featuresB)

        inner_loop_nsamples = int(round(np.sqrt(num_samples)))

        samples_input, samples_delta = self._get_samples_input(input_tensor, baseline,
                                                               inner_loop_nsamples, use_expectation)

        interaction_mult_tensor = torch.zeros([input_tensor.size(0), samples_input.size(1),
                                               len(featuresA), len(featuresB)])

        ig_tensor = torch.zeros(samples_input.shape[:-1] + (len(features),)).float()

        if use_expectation:
            loop_num = inner_loop_nsamples
        else:
            loop_num = inner_loop_nsamples + 1

        if verbose:
            iterable = tqdm(range(loop_num))
        else:
            iterable = range(loop_num)

        for i in iterable:
            particular_slice = samples_input[:,i]
            self.model_ = self.model
            detached_slice = particular_slice.clone().detach()

            def temp_model(inputs):
                detached_slice[:,features] = inputs
                return self.model_(detached_slice)

            self.model = temp_model
            particular_inputs = particular_slice[:,features]
            ig_tensor[:,i,:] = self.attributions(
                particular_inputs,
                baseline[:,features],
                num_samples=inner_loop_nsamples,
                use_expectation=use_expectation,
                output_indices=output_indices
            )
            self.model = self.model_

            for feature in range(len(featuresA)):
                second_grads = grad(
                    outputs=ig_tensor[:,i,feature],
                    inputs=particular_inputs,
                    grad_outputs=torch.ones_like(ig_tensor[:,i,feature]),
                    create_graph=True)[0].detach()
                interaction_mult_tensor[:,i,feature,:] = second_grads[:,len(featuresA):]

        interaction_mult_tensor = interaction_mult_tensor.to(samples_delta.device)
        interaction_tensor = interaction_mult_tensor * samples_delta[..., featuresB].unsqueeze(2)
        interactions = interaction_tensor.mean(1)

        return interactions



from chemprop.features.featurization import MolGraph, BatchMolGraph
from chemprop.models import MoleculeModel
from .explained_mol_graph import ExplainedMolGraph
from torch.nn.functional import softmax
from lime.lime_tabular import LimeTabularExplainer
from copy import deepcopy
from torch import Tensor
from torch.nn import functional
import numpy as np


class LIMEExplainer:
    # +1 is to include room for common values, as mentioned in chemprop.features.featurization, line 31
    feature_sizes = {
        "atom_type": 100+1,
        "#bonds": 6+1,
        "formal_charge": 5+1,
        "chirality": 4+1,
        "#Hs": 5+1,
        "hybridization": 5+1
    }.values()

    possible_atoms = [[i%101, i%7, i%6, i%5, i%6, i%6, i%2, i/100] for i in range(101)]
    possible_bonds = [[i%5, i%2, i%2, i%7] for i in range(7)]

    def __init__(self, model: MoleculeModel, mol: str=None):
        self.model = model
        self.model.eval() # Trigger evaluation mode
        if mol is not None:
            self._set_molecule(mol)
        else:
            self.mol = None

    def _set_molecule(self, mol: str):
        mol = MolGraph(mol)
        self.mol = mol
        input_features = []
        self.categorical = []
        self.atom_slices = [0]
        values = []
        i = 0
        # atoms to array
        for f_atom in mol.f_atoms:
            values = []
            j = 0
            for size in self.feature_sizes:
                values.append(f_atom[j:j+size].index(1))
                j += size
            values.extend(f_atom[j:])
            input_features += values
            self.categorical += list(range(i, i+len(values)-1))
            i += len(values)
            self.atom_slices.append(i)
        self.atom_size = len(values)
        # bonds to array
        self.bond_slices = [i]
        for f_bond in mol.f_bonds:
            f_bond = f_bond[-14:]
            values = []
            values.append(f_bond[:5].index(1))  # bond type
            values.extend(f_bond[5:7])          # conjugated / in ring
            values.append(f_bond[7:].index(1))  # stereo (none, any, E/Z, or cis/trans)
            input_features += values
            self.categorical += list(range(i, i+4))
            i += 4
            self.bond_slices.append(i)
        self.bond_size = len(values)
        self.input = np.array(input_features, dtype="float32")

    def _array_to_mol(self, array) -> MolGraph:
        # parsing atoms features
        f_atoms = []
        for i, j in zip(self.atom_slices[:-1], self.atom_slices[1:]):
            values = array[i:j]
            f_atom = []
            for k, size in enumerate(self.feature_sizes):
                feature = [0]*size
                feature[int(values[k])] = 1
                f_atom += feature
            f_atom.extend(values[-2:])
            f_atoms.append(f_atom)
        # parsing bonds features (and include atoms features)
        f_bonds = []
        bounds = zip(self.bond_slices[:-1], self.bond_slices[1:])
        for a, (i, j) in zip(self.mol.b2a, bounds):
            values = array[i:j]
            f_bond = [0]*14
            f_bond[int(values[0])] = 1
            f_bond[5:7] = values[1:3]
            f_bond[7 + int(values[3])] = 1
            f_bonds.append(np.concatenate((f_atoms[a], f_bond)))
        # creating a copy of the MolGraph and change its features
        mol = deepcopy(self.mol)
        mol.f_atoms = f_atoms
        mol.f_bonds = f_bonds
        return mol

    def predict_proba(self, inputs, softmax_output=False):
        """Compute the model's prediction for arrays inputs

        inputs: list of arrays, containing features for the current molecule
        softmax_output: flag to apply a softmax. If False the original linear output is returned

        Returns a numpy array of predictions, of shape (len(inputs), nb_classes)
        """
        if self.mol is None:
            raise AttributeError("You must set a molecule first (set_molecule method).")
        # converting arrays to MolGraphs
        mols = list(map(self._array_to_mol, inputs))
        # getting the model predictions for the molecules
        output = self.model([BatchMolGraph(mols)]).detach()
        output = Tensor.cpu(output)
        #if softmax_output:
        #    output = softmax(output, dim=1)
        #else:
        #    output=functional.normalize(output, p=1, dim=1)
        return output.numpy()

    def _select_class(self, predictions, n):
        """
        predictions: array of shape (_, nb_classes)
            (warning: coefficients of this array MUST be between 0 and 1)
        n: class to select

        Return an array of shape (_, 2):
            - first column corresponding to 1 - the class output
            - second column corresponding to the class output
        """
        class_output = predictions[:,n:n+1]
        return np.concatenate([1-class_output, class_output], axis=1)

    def explain_molecule(self, mol: str, n: int = 1) -> ExplainedMolGraph:
        """
        mol: molecule in SMILES format
        n: class number to explain (starting at 0)
        """
        self._set_molecule(mol)
        # Initializing the explainer
        dataset = []
        for i in range(101):
            instance = []
            for j in range(self.mol.n_atoms):
                instance += self.possible_atoms[(i+j)%101]
            for j in range(self.mol.n_bonds):
                instance += self.possible_bonds[(i+j)%7]
            dataset.append(instance)
        explainer = LimeTabularExplainer(np.array(dataset), categorical_features=self.categorical,
                                         discretize_continuous=False)
        # Getting the mean local contribution for each feature
        contribs = np.zeros(len(self.input))
        for _ in range(5):
            exp = explainer.explain_instance(self.input, lambda x: self._select_class(self.predict_proba(x), n),
                                             num_features=len(self.input), num_samples=1000)
            for i, value in exp.local_exp[1]:
                contribs[i] += value / 5
        # Making an ExplainedGraphMol and returning it
        f_atoms = contribs[:self.mol.n_atoms*self.atom_size].reshape((-1, self.atom_size))
        f_bonds = contribs[self.mol.n_atoms*self.atom_size:].reshape((-1, self.bond_size))
        exp_mol = ExplainedMolGraph(mol, f_atoms, f_bonds)
        return exp_mol

    def __call__(self, *args, **kwargs):
        return self.predict_proba(*args, **kwargs)



from flask import Flask, render_template, request, redirect

from .chemprop_lime import *
from .chemprop_path_explain import *
from .explainer_utils import *

from chemprop.utils import load_checkpoint

from matplotlib.pyplot import cm
import os
import hashlib


def mkdir(path):
    if not os.path.isdir(path):
        os.makedirs(path)
        return True
    return False


cmap = cm.get_cmap("bone_r")
color = lambda x: tuple(map(lambda x: int(x*255), cmap(x)[:3]))
explainers = list(enumerate(["PathExplain", "LIME"]))
exps_ids = ["path_explain", "lime"]

tmp_path = "/tmp/chemexp_data/"

app = Flask(__name__)


class GUIServer:
    def __init__(self):
        self.models_dir = None
        self.checkpoints = None
        self.models = None
        self.model_names = None
        
        # creating temporary folder (for generated figures)
        mkdir(tmp_path)
        self.cd = "/".join(__file__.split("/")[:-1])
        if not os.path.islink(self.cd + "/static/tmp"):
            os.symlink(tmp_path, self.cd + "/static/tmp")
        
    
    def set_models_dir(self, models_dir):
        """
        models_dir: folder containing checkpoints generated by chemprop (.pt files)
        """
        models_dir += "/"
        self.models_dir = models_dir
        self.checkpoints = {
            file.rstrip(".pt"): load_checkpoint(models_dir+file)
            for file in os.listdir(models_dir)
            if file.endswith(".pt") and os.path.isfile(models_dir+file)
        }
        self.models = [
            {
                "name": name,
                "lime": LIMEExplainer(self.checkpoints[name]),
                "path_explain": PathExplainer(self.checkpoints[name])
            }
            for name in self.checkpoints
        ]
        self.model_names = list(enumerate([m["name"] for m in self.models]))

    
    def run(self, **kwargs):
        """
        **kwargs: arguments for the Flask server
        """
        app.run(**kwargs)


@app.route("/")
def index():
    return render_template("index.html", models=server.model_names)


@app.route("/display", methods=["POST"])
def display():
    mol = request.form["smiles"]
    try:
        model = server.models[int(request.form["model"])]["lime"]
    except:
        model = server.models[0]["lime"]
    # image generation
    mkdir(tmp_path + "mols")
    mol_md5 = hashlib.md5(mol.encode()).hexdigest()
    mol_to_png(mol, tmp_path + "mols/" + mol_md5 + ".png", use_color=True)
    # model prediction
    try:
        model._set_molecule(mol)
        predictions = model.predict_proba([model.input])[0]
        n_atoms = model.mol.n_atoms
        n_bonds = model.mol.n_bonds
    except:
        predictions = [0]
        n_atoms = 0
        n_bonds = 0
    return render_template("display.html", mol_md5=mol_md5,
                           explainers=explainers,
                           predictions=predictions,
                           predcolors=[color(x) for x in predictions],
                           n_predictions=len(predictions),
                           n_atoms=n_atoms,
                           n_bonds=n_bonds)


@app.route("/explain", methods=["POST"])
def explain():
    classNb = int(request.form["class"])
    mol = request.form["smiles"]
    try:
        explainer = exps_ids[int(request.form["exp"])]
        model = server.models[int(request.form["model"])]
    except:
        explainer = exps_ids[0]
        model = server.models[0]
    print(mol)
    # create an explanation and save it in a json file,
    # or load it if it already exists
    base_dir = f"{server.cd}/static/{model['name']}/{explainer}"
    mkdir(base_dir)
    mol_hash = hashlib.md5(mol.encode()).hexdigest()
    filename = f"{base_dir}/{mol_hash}_{classNb}.json"
    if not os.path.isfile(filename):
        if explainer == "lime":
            exp = model["lime"].explain_molecule(mol, classNb)
        else:
            exp = model["path_explain"].explain_molecule(mol, classNb)
        exp_to_json(exp, filename)
    else:
        exp = json_to_exp(filename)
    # generate images
    static_path = f"static/tmp/{model['name']}/{explainer}/{mol_hash}_{classNb}"
    mkdir(f"{server.cd}/{static_path}")
    exp_to_png(exp, f"{server.cd}/{static_path}/avg.png")
    for i in range(8):
        atoms_exp_to_png(exp, i, f"{server.cd}/{static_path}/exp_atom_{i}.png")
    for i in range(4):
        bonds_exp_to_png(exp, i, f"{server.cd}/{static_path}/exp_bond_{i}.png")
    bonds_max = abs(exp.f_bonds).max() if len(exp.f_bonds) > 0 else 0
    atoms_max = abs(exp.f_atoms).max() if len(exp.f_atoms) > 0 else 1
    bound = round(max(bonds_max, atoms_max), 3)

    return render_template("results.html",
                           mol=mol,
                           path=static_path,
                           bound=bound,
                           classNb=classNb,
                           explainer=explainer)


@app.route("/interactions", methods=["POST"])
def interactions():
    classNb = int(request.form["class"])
    mol = request.form["smiles"]
    try:
        feature1 = int(request.form["feature1"])
        feature2 = int(request.form["feature2"])
        model = server.models[int(request.form["model"])]
    except:
        feature1 = 0
        feature2 = 5
        model = server.models[0]
    print(mol)
    # create an explanation and save it in a json file,
    # or load it if it already exists
    mkdir(f"{server.cd}/static/{model['name']}/path_explain/interactions")
    mol_hash = hashlib.md5(mol.encode()).hexdigest()
    filename = f"{server.cd}/static/{model['name']}/path_explain/interactions/{mol_hash}_{classNb}_{feature1}_{feature2}.json"
    if not os.path.isfile(filename):
        exp = model["path_explain"].explain_interactions(mol, feature1, feature2, classNb)
        exp_to_json(exp, filename)
    else:
        exp = json_to_exp(filename)
    # generate images
    static_path = "static/tmp/path_explain" + f"/{mol_hash}_{model['name']}_{classNb}"
    mkdir(f"{server.cd}/{static_path}")
    img_name = f"{static_path}/interactions_{classNb}_{feature1}_{feature2}.png"
    atoms_exp_to_png(exp, 0, f"{server.cd}/{img_name}")
    atoms_max = abs(exp.f_atoms).max() if len(exp.f_atoms) > 0 else 1
    bound = round(atoms_max, 3)

    return render_template("interactions.html",
                           path=img_name,
                           bound=bound,
                           classnb=classNb)


server = GUIServer()



from chemprop.features.featurization import MolGraph
import numpy as np

class ExplainedMolGraph(MolGraph):
    def __init__(self, smiles, f_atoms, f_bonds):
        super(ExplainedMolGraph, self).__init__(smiles)
        self.f_atoms = np.array(f_atoms)
        self.f_bonds = np.array(f_bonds)
        self.smiles = smiles



from .explained_mol_graph import ExplainedMolGraph
from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D
from matplotlib.pyplot import cm
import numpy as np
import shutil
import json


def exp_to_json(exp: ExplainedMolGraph, filename: str):
    with open(filename, 'w') as f:
        json.dump({
            "a2b": exp.a2b,
            "b2a": exp.b2a,
            "b2revb": exp.b2revb,
            "f_atoms": exp.f_atoms.tolist(),
            "f_bonds": exp.f_bonds.tolist(),
            "n_atoms": exp.n_atoms,
            "n_bonds": exp.n_bonds,
            "smiles": exp.smiles
        }, f)


def json_to_exp(filename: str):
    with open(filename, 'r') as f:
        exp = json.load(f)
    return ExplainedMolGraph(exp["smiles"], exp["f_atoms"], exp["f_bonds"])


def exp_to_png(exp: ExplainedMolGraph, filename: str, cmap="bwr", size=(1200, 675)):
    # define a color map
    cmap = cm.get_cmap(cmap)
    color = lambda x: cmap(x)[:3]
    # get the corresponding molecule with RDKit
    mol = Chem.MolFromSmiles(exp.smiles)
    # create a dictionary Atom id --> Mean contribution
    atom_colors = dict()
    for i, e in enumerate(exp.f_atoms):
        atom_colors[i] = np.mean(e)

    # create a dictionary Bond id --> Mean contribution
    # (however the bonds are oriented, so we take the average of the two directions)
    bond_colors = dict()
    i = 0
    for a1 in range(exp.n_atoms):
        for a2 in range(a1 + 1, exp.n_atoms):
            bond = mol.GetBondBetweenAtoms(a1, a2)
            if bond is None:
                continue
            e = np.mean(np.concatenate((exp.f_bonds[i], exp.f_bonds[i + 1])))
            bond_colors[bond.GetIdx()] = e
            i += 2

    # convert mean contributions to colors
    values = list(atom_colors.values()) + list(bond_colors.values())
    maxi = max(abs(min(values)), abs(max(values)))
    for colors in atom_colors, bond_colors:
        for k in colors:
            colors[k] = color(colors[k]/maxi + 0.5)

    # draw the molecule with RDKit
    d = rdMolDraw2D.MolDraw2DCairo(*size)
    options = d.drawOptions()
    options.setAtomPalette({})
    d.SetDrawOptions(options)
    rdMolDraw2D.PrepareAndDrawMolecule(d, mol,
                                       highlightAtoms=list(range(exp.n_atoms)),
                                       highlightAtomColors=atom_colors,
                                       highlightBonds=list(range(exp.n_bonds // 2)),
                                       highlightBondColors=bond_colors)
    # save the figure
    d.WriteDrawingText(filename)
    print("Figure saved to " + filename)


def atoms_exp_to_png(exp: ExplainedMolGraph, feature: int, filename: str, cmap="bwr", size=(1200, 675)):
    # define a color map
    cmap = cm.get_cmap(cmap)
    color = lambda x: cmap(x)[:3]
    # get the corresponding molecule with RDKit
    mol = Chem.MolFromSmiles(exp.smiles)
    # create a dictionary Atom id --> Mean contribution
    atom_colors = dict()
    bonds_max = abs(exp.f_bonds).max() if len(exp.f_bonds) > 0 else 0
    atoms_max = abs(exp.f_atoms).max() if len(exp.f_atoms) > 0 else 1
    max_exp = max(bonds_max, atoms_max) * 2
    if max_exp == 0: max_exp = 1
    for i, e in enumerate(exp.f_atoms):
        atom_colors[i] = e[feature] / max_exp

    # convert mean contributions to colors
    for k in atom_colors:
        atom_colors[k] = color(atom_colors[k] + 0.5)

    # draw the molecule with RDKit
    d = rdMolDraw2D.MolDraw2DCairo(*size)
    options = d.drawOptions()
    options.setAtomPalette({})
    d.SetDrawOptions(options)
    rdMolDraw2D.PrepareAndDrawMolecule(d, mol,
                                       highlightAtoms=list(range(exp.n_atoms)),
                                       highlightAtomColors=atom_colors)
    # save the figure
    d.WriteDrawingText(filename)
    print("Figure saved to " + filename)


def bonds_exp_to_png(exp: ExplainedMolGraph, feature: int, filename: str, cmap="bwr", size=(1200, 675)):
    # define a color map
    cmap = cm.get_cmap(cmap)
    color = lambda x: cmap(x)[:3]
    # get the corresponding molecule with RDKit
    mol = Chem.MolFromSmiles(exp.smiles)

    # create a dictionary Bond id --> Mean contribution
    # (however the bonds are oriented, so we take the average of the two directions)
    bond_colors = dict()
    bonds_max = abs(exp.f_bonds).max() if len(exp.f_bonds) > 0 else 0
    atoms_max = abs(exp.f_atoms).max() if len(exp.f_atoms) > 0 else 1
    max_exp = max(bonds_max, atoms_max) * 2
    if max_exp == 0: max_exp = 1
    i = 0
    for a1 in range(exp.n_atoms):
        for a2 in range(a1 + 1, exp.n_atoms):
            bond = mol.GetBondBetweenAtoms(a1, a2)
            if bond is None:
                continue
            e = (exp.f_bonds[i][feature] + exp.f_bonds[i + 1][feature]) / 2
            bond_colors[bond.GetIdx()] = e / max_exp
            i += 2

    # convert mean contributions to colors
    for k in bond_colors:
        bond_colors[k] = color(bond_colors[k] + 0.5)

    # draw the molecule with RDKit
    d = rdMolDraw2D.MolDraw2DCairo(*size)
    options = d.drawOptions()
    options.setAtomPalette({})
    d.SetDrawOptions(options)
    rdMolDraw2D.PrepareAndDrawMolecule(d, mol,
                                       highlightBonds=list(range(exp.n_bonds // 2)),
                                       highlightBondColors=bond_colors)
    # save the figure
    d.WriteDrawingText(filename)
    print("Figure saved to " + filename)


def mol_to_png(mol, filename: str, size=(1200, 675), use_color=False):
    # get the corresponding molecule with RDKit
    try:
        mol = Chem.MolFromSmiles(mol)
    except:
        mol = None
    if mol is None:
        cd = "/".join(__file__.split("/")[:-1])
        if size[0]/size[1] < 1.4:
            shutil.copyfile(cd + "/static/smiles_error.png", filename)
        else:
            shutil.copyfile(cd + "/static/smiles_error_rect.png", filename)
    else:
        # draw the molecule with RDKit
        d = rdMolDraw2D.MolDraw2DCairo(*size)
        if not use_color:
            options = d.drawOptions()
            options.setAtomPalette({})
            d.SetDrawOptions(options)
        rdMolDraw2D.PrepareAndDrawMolecule(d, mol)
        # save the figure
        d.WriteDrawingText(filename)



import pathlib
from distutils.core import setup


HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
  name = "chemexp",
  packages = ["chemexp"],
  version = "0.2.3",
  description = "An explainer for Chemprop based on LIME and PathExplain.",
  long_description = README,
  long_description_content_type = "text/markdown",
  author = "Fabien BERNIER",
  author_email = "fabien.bernier@telecomnancy.net",
  url = "https://github.com/Locust2520/chemexp",
  keywords = [
    "explainer",
    "explanation",
    "visualization",
    "chemprop",
    "molecule",
    "model",
    "deep learning",
    "graph"
  ],
  classifiers = [
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.8"
  ],
)
